import { MessageTypes } from "../enums/MessageTypes.mjs";
import { DefaultOptions } from "../options/defaults/DefaultOptions.mjs";
import { DefaultSubtitlesSettings } from "../options/defaults/DefaultSubtitlesSettings.mjs";
import { EnvUtils } from "./EnvUtils.mjs";

const DAILY_DOWNLOAD_LIMIT = 3;
const DOWNLOAD_COUNT_KEY = "daily_download_count";
const LAST_DOWNLOAD_DATE_KEY = "last_download_date";

export class Utils {
    static async checkDownloadLimit() {
        const today = new Date().toDateString();
        const lastDate = await Utils.getConfig(LAST_DOWNLOAD_DATE_KEY);
        const downloadCount = await Utils.getConfig(DOWNLOAD_COUNT_KEY) || 0;

        if (lastDate !== today) {
            await Utils.setConfig(LAST_DOWNLOAD_DATE_KEY, today);
            await Utils.setConfig(DOWNLOAD_COUNT_KEY, 0);
            return true;
        }

        return downloadCount < DAILY_DOWNLOAD_LIMIT;
    }

    static async incrementDownloadCount() {
        const count = (await Utils.getConfig(DOWNLOAD_COUNT_KEY) || 0) + 1;
        await Utils.setConfig(DOWNLOAD_COUNT_KEY, count);
        return count;
    }
    static getOptionsFromStorage() {
        return Utils.loadAndParseOptions("options", DefaultOptions);
    }

    static getSubtitlesSettingsFromStorage() {
        return Utils.loadAndParseOptions(
            "subtitlesSettings",
            DefaultSubtitlesSettings
        );
    }

    static mergeOptions(defaultOptions, newOptions) {
        const options = {};
        for (const prop in defaultOptions) {
            if (Object.hasOwn(defaultOptions, prop)) {
                const opt = defaultOptions[prop];
                if (typeof opt === "object" && !Array.isArray(opt)) {
                    options[prop] = this.mergeOptions(
                        opt,
                        newOptions[prop] || {}
                    );
                } else {
                    options[prop] =
                        Object.hasOwn(newOptions, prop) &&
                        typeof newOptions[prop] === typeof opt
                            ? newOptions[prop]
                            : opt;
                }
            }
        }
        return options;
    }

    /**
     * Binary search utility.
     * @param {array} array
     * @param {*} el
     * @param {function} compareFn
     * @return {*}
     */
    static binarySearch(array, el, compareFn) {
        let lower = 0;
        let upper = array.length - 1;
        while (lower <= upper) {
            const middle = (upper + lower) >> 1;
            const cmp = compareFn(el, array[middle]);
            if (cmp > 0) {
                lower = middle + 1;
            } else if (cmp < 0) {
                upper = middle - 1;
            } else {
                return middle;
            }
        }
        return -lower - 1;
    }

    static clamp(value, min, max) {
        return Math.min(Math.max(value, min), max);
    }

    static getDataByteSize(data) {
        if (typeof data === "string") return data.length * 2;
        if (data instanceof ArrayBuffer) return data.byteLength;
        if (data instanceof Blob) return data.size;
        return 0;
    }

    static zipTimedFragments(tracks) {
        const zippedFragments = [];
        tracks.forEach((fragments, track) => {
            fragments.forEach((fragment) => {
                zippedFragments.push({
                    track,
                    fragment,
                });
            });
        });

        zippedFragments.sort((a, b) => {
            return a.fragment.start - b.fragment.start;
        });

        return zippedFragments;
    }

    static async loadAndParseOptions(key, defaultOptions) {
        const settingsStr = await Utils.getConfig(key);
        if (settingsStr) {
            try {
                const settings = JSON.parse(settingsStr);
                return Utils.mergeOptions(defaultOptions, settings);
            } catch (e) {
                console.error(e);
            }
        }
        return Utils.mergeOptions(defaultOptions, {});
    }

    static getConfig(key) {
        return new Promise((resolve, reject) => {
            if (EnvUtils.isExtension()) {
                chrome.storage.local.get(key, (result) => {
                    resolve(result[key]);
                });
            } else {
                resolve(localStorage.getItem(key));
            }
        });
    }

    static setConfig(key, value) {
        return new Promise((resolve, reject) => {
            if (EnvUtils.isExtension()) {
                chrome.storage.local.set({ [key]: value }, () => {
                    resolve();
                });
            } else {
                localStorage.setItem(key, value);
                resolve();
            }
        });
    }

    static selectQuality(levels, defaultQuality) {
        let max = -1;
        let maxLevel = undefined;
        let min = Number.MAX_SAFE_INTEGER;
        let minLevel = undefined;

        if (defaultQuality === "Auto") {
            const qualityMultiplier = 1.1;

            // Get best quality but within screen resolution
            levels.forEach((level, key) => {
                if (level.bitrate > max) {
                    if (
                        level.width >
                            window.innerWidth *
                                window.devicePixelRatio *
                                qualityMultiplier &&
                        level.height >
                            window.innerHeight *
                                window.devicePixelRatio *
                                qualityMultiplier
                    ) {
                    } else {
                        max = level.bitrate;
                        maxLevel = key;
                    }
                }

                if (level.bitrate < min) {
                    min = level.bitrate;
                    minLevel = key;
                }
            });

            if (maxLevel === undefined) {
                maxLevel = minLevel;
            }

            return maxLevel;
        } else {
            const desiredHeight = parseInt(defaultQuality.replace("p", ""));
            const list = [];
            levels.forEach((level, key) => {
                list.push({
                    key,
                    level,
                    diff: Math.abs(level.height - desiredHeight),
                });
            });

            // Sort by height difference and then by bitrate. Choose highest bitrate if multiple have the same height difference
            list.sort((a, b) => {
                if (a.diff === b.diff) {
                    return a.level.bitrate - b.level.bitrate;
                }
                return a.diff - b.diff;
            });

            return list[0]?.key;
        }
    }

    static timeoutableEvent(context, event, timeout) {
        let callback;
        let timeoutId;
        const cleanup = () => {
            clearTimeout(timeoutId);
            context.off(event, callback);
        };

        return new Promise((resolve, reject) => {
            callback = () => {
                cleanup();
                resolve(true);
            };

            timeoutId = setTimeout(() => {
                cleanup();
                resolve(false);
            }, timeout);

            context.on(event, callback);
        });
    }

    static asyncTimeout(timeout) {
        return new Promise((resolve, reject) => {
            setTimeout(() => {
                resolve();
            }, timeout);
        });
    }

    static async downloadURL(url, filename, forceDirect = false) {
        // Firefox has a bug where it doesn't download filed from sandboxed iframes
        // Caused by bloburl partitioning issues. See gecko's dom/file/uri/BlobURLProtocolHandler.cpp#L775C1-L786C6
        if (EnvUtils.isExtension() && !EnvUtils.isChrome() && !forceDirect) {
            return new Promise((resolve, reject) => {
                chrome.runtime.sendMessage(
                    {
                        type: MessageTypes.DOWNLOAD,
                        url,
                        filename,
                    },
                    (response) => {
                        resolve(response);
                    }
                );
            });
        } else {
            const aElement = document.createElement("a");
            aElement.href = url;
            aElement.download = filename;
            aElement.target = "_blank";
            document.body.appendChild(aElement);
            aElement.click();
            aElement.remove();
        }
    }

    static findPropertyRecursive(obj, key, list = [], stack = []) {
        if (typeof obj !== "object" || obj === null) {
            return;
        }

        if (Array.isArray(obj)) {
            obj.forEach((v, i) => {
                stack.push(i);
                Utils.findPropertyRecursive(v, key, list, stack);
                stack.pop();
            });
        } else {
            if (Object.hasOwn(obj, key)) {
                list.push({ value: obj[key], stack: stack.slice(), obj });
            }
            Object.keys(obj).forEach((k) => {
                stack.push(k);
                Utils.findPropertyRecursive(obj[k], key, list, stack);
                stack.pop();
            });
        }

        return list;
    }
}
