import { MessageTypes } from "../player/enums/MessageTypes.mjs";
import { Utils } from "../player/utils/Utils.mjs";
import { gtagEvent } from "../gtagEvent.mjs";


let popupStatus = "CLOSED";

// DOM元素引用
const DOM = {
    result: document.getElementsByClassName("result"),
    thumbnail: document.querySelector(".result .result-info img"),
    title: document.querySelector(".result .result-info .title"),
    loading: document.getElementsByClassName("loading"),
    product: document.querySelector(".content .product"),
    productDesc: document.querySelector(".content .product-desc"),
    productBtn: document.querySelector(".content .product-btn"),
    blankTip: document.getElementsByClassName("blank-tip"),
    limitTip: document.getElementsByClassName("result-tip"),
    limitTipContent: document.querySelector(".result-tip .tip-content"),
    limitTipBtn: document.querySelector(".result-tip .tip-btn")
};

// 流媒体数据存储
const mediaStreams = {
    video: new Map(),
    audio: new Map(),
    videoLastKey: null,
    audioLastKey: null,
};

(() => {
    initializePopup();
    initializeProductBtn();
    setupVisibilityListener();
    setupMessageListener();
})();

// 初始化弹窗
function initializePopup() {
    popupStatus = "LODDING";
    chrome.runtime.sendMessage({
        type: MessageTypes.POPUP_LOADED,
    });
    gtagEvent("total_user");
    chrome.storage.local.get("FIRST_INSTALL", (result) => {
        if (!result || !result.FIRST_INSTALL) {
            gtagEvent("new_user");
            chrome.storage.local.set({
                FIRST_INSTALL: true,
            });
        }
    });
}
// 初始化产品按钮
function initializeProductBtn() {
    DOM.productBtn.addEventListener("click", () => {
        gtagEvent(popupStatus === "LODDING" ? "desktop_click_loading" : "clicdesktop_click_no_resultk_try");
        window.open("https://anyukit.com/youtube-downloader/?td=extension", "_blank");
    });
}
// 设置可见性监听器
function setupVisibilityListener() {
    document.addEventListener("visibilitychange", () => {
        if (document.visibilityState === "hidden") {
            popupStatus = "CLOSED";
            chrome.runtime.sendMessage({ type: MessageTypes.POPUP_UNLOADED });
        }
    });
}
// 设置消息监听器
function setupMessageListener() {
    chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
        switch (request.type) {
            case MessageTypes.NO_VIDEOS:
                handleNoVideos(sendResponse);
                break;
            case MessageTypes.ALREADY_INFO:
                handleMediaInfo(request.info);
                break;
        }
    });
}
// 处理下载限制
async function handleLimitDownload() {
    if (mediaStreams.duration > 600) {
        DOM.limitTipBtn.addEventListener("click", () => {
            gtagEvent('desktop_click_10min_limit')
            window.open("https://anyukit.com/youtube-downloader/?td=extension", "_blank");
        });
        DOM.limitTipContent.innerHTML = `To download long videos & audio <strong>over 10 minutes</strong>, try the desktop version.`;
        DOM.limitTip[0].style.display = "flex";
        
        return false;
    } else if (!(await Utils.checkDownloadLimit())) {
        DOM.limitTipBtn.addEventListener("click", () => {
            gtagEvent('desktop_click_daily_limit')
            window.open("https://anyukit.com/youtube-downloader/?td=extension", "_blank");
        });
        DOM.limitTipContent.innerHTML = `<strong>Daily 3-download limit</strong> reached. Try the desktop version for unlimited downloads.`;
        DOM.limitTip[0].style.display = "flex";
        
        return false;
    }
    return true;
}

// 处理无视频情况
function handleNoVideos(sendResponse) {
    popupStatus = "NO_VIDEOS";
    DOM.loading[0].style.display = "none";
    DOM.blankTip[0].style.display = "block";
    DOM.productDesc.innerHTML =
        "Restart or try the desktop version for a more stable search experience.";
    gtagEvent("show_no_result");
    sendResponse("ok");
}

// 处理媒体信息
function handleMediaInfo(info) {
    popupStatus = "ALREADY_INFO";
    const { basic_info, streaming_data } = info;
    processStreams(streaming_data.adaptive_formats);
    mediaStreams.duration = basic_info.duration;
    updateUIWithBasicInfo(basic_info);
    renderMediaLists();
    addResultListener();
    gtagEvent("show_result");
}

// 处理流数据
function processStreams(formats) {
    formats.forEach((stream) => {
        if (
            stream.mime_type.includes("video/mp4") &&
            ["480p", "360p", "240p", "144p"].includes(stream.quality_label)
        ) {
            mediaStreams.video.set(stream.quality_label, stream);
        } else if (
            stream.mime_type.includes("audio/") &&
            stream.bitrate < 140000
        ) {
            mediaStreams.audio.set(stream.bitrate, stream);
        }
    });
    if (mediaStreams.video.size > 3) {
        mediaStreams.video.delete("144p");
    }
    if (mediaStreams.audio.size > 1) {
        mediaStreams.audio = new Map([[...mediaStreams.audio].pop()]);
    }

    mediaStreams.videoLastKey = Array.from(mediaStreams.video.keys()).pop();
    mediaStreams.audioLastKey = Array.from(mediaStreams.audio.keys()).pop();
}

// 更新基本信息UI
function updateUIWithBasicInfo(basicInfo) {
    DOM.loading[0].style.display = "none";
    DOM.product.style.display = "none";
    DOM.result[0].style.display = "block";
    DOM.thumbnail.src = basicInfo.thumbnail[0].url;
    DOM.title.innerHTML = basicInfo.title;
}

// 渲染媒体列表
function renderMediaLists() {
    if (mediaStreams.video.size) {
        renderMediaList("video");
    }
    if (mediaStreams.audio.size) {
        renderMediaList("audio");
    }
}

// 创建媒体列表
function renderMediaList(type) {
    const list = document.createElement("div");
    const listLabel = document.createElement("div");
    list.className = "result-list";
    listLabel.className = "result-list-label";
    listLabel.innerHTML = type === "video" ? "Video" : "Music";
    list.appendChild(listLabel);
    if (type === "video") {
        list.appendChild(createItem({}, "video", "1080p", false));
        list.appendChild(createItem({}, "video", "720p", false));
    } else {
        list.appendChild(createItem({}, "audio", 320000, false));
        list.appendChild(createItem({}, "audio", 192000, false));
    }
    mediaStreams[type].forEach((stream, quality) => {
        list.appendChild(createItem(stream, type, quality, true));
    });

    DOM.result[0].appendChild(list);
}

// 创建项
function createItem(stream, type, quality, saveable) {
    const item = document.createElement("div");
    item.className = "result-list-item";
    item.innerHTML = `
        <div class="quality">
            <span ${
                saveable
                    ? ""
                    : type === "video"
                    ? 'class="quality-hd'
                    : 'class="quality-hq'
            }">${type === "video" ? quality : Math.floor(quality / 1000) + 'k'}</span>
        </div> 
        <div class="format">${type === "video" ? "MP4" : "MP3"}</div>
        ${
            saveable
                ? `<div class="download-btn"><div class="download-btn_cansave download-btn_${type}" data-itag="${stream.itag}">Download</div></div>`
                : `<a class="download-btn" data-quality="${type === "video" ? quality : Math.floor(quality / 1000) + 'k'}">Download in Desktop ></a>`
        }
    `;
    return item;
}

// 添加结果监听器
function addResultListener() {
    setupVideoDownload();
    setupDesktopDownload();
    setupAudioDownload();
}

// 设置视频下载
function setupVideoDownload() {
    const downloadBtnVideos = document.querySelectorAll(".download-btn_video");
    downloadBtnVideos.forEach((downloadBtnVideo) => {
        let itag = downloadBtnVideo.getAttribute("data-itag");
        downloadBtnVideo.addEventListener("click", async () => {
            if (!(await handleLimitDownload())) {
                return;
            }
            downloadBtnVideo.classList.add("download-btn_loading");
            gtagEvent('dl_downloads')
            gtagEvent('dl_downloads_mp4')
            chrome.runtime.sendMessage({
                type: MessageTypes.DOWNLOAD_VIDEO,
                itag,
            });
        });
    });
}

// 设置桌面版下载
function setupDesktopDownload() {
    const downloadBtns = document.querySelectorAll("a.download-btn");
    downloadBtns.forEach((btn) => {
        btn.addEventListener("click", () => {
            gtagEvent(`desktop_click_${btn.dataset.quality}`)
            chrome.runtime.sendMessage({ type: MessageTypes.POPUP_UNLOADED });
            window.open(
                "https://anyukit.com/youtube-downloader/?td=extension",
                "_blank"
            );
        });
    });
}

// 设置音频下载
function setupAudioDownload() {
    const downloadBtnAudio = document.querySelector(".download-btn_audio");

    if (downloadBtnAudio && mediaStreams.audioLastKey) {
        downloadBtnAudio.addEventListener("click", async () => {
            if (!(await handleLimitDownload())) {
                return;
            }
            downloadBtnAudio.classList.add("download-btn_loading");
            gtagEvent('dl_downloads')
            gtagEvent('dl_downloads_mp3')
            downloadAudioWithFetch(
                mediaStreams.audio.get(mediaStreams.audioLastKey).url,
                "audio.mp3"
            );
        });
    }
}

// 下载音频
async function downloadAudioWithFetch(url, filename) {
    try {
        const response = await fetch(url);
        if (!response.ok) {
            throw new Error(`HTTP error! status: ${response.status}`);
        }
        const blob = await response.blob();
        const blobUrl = window.URL.createObjectURL(blob);
        const a = document.createElement("a");
        a.href = blobUrl;
        a.download = filename;
        document.body.appendChild(a);
        a.click();

        // 清理资源
        window.URL.revokeObjectURL(blobUrl);
        document.body.removeChild(a);
        await Utils.incrementDownloadCount();
    } catch (error) {
        console.error("音频下载失败:", error.message);
    }
}
